#ifndef PRIMITIVE_H_
#define PRIMITIVE_H_

#include <Eigen/Core>
#include <set>
#include <vector>

// Region signature structure that contains
//    covariance - Matrix C of the region.
//    null_space - Null space of matrix C.
//    num_points - Number of points in the region.
struct Signature
{
	Eigen::MatrixXd covariance;
	Eigen::MatrixXd null_space;
	int num_points;
};

// Compute the signature of the region.
// The input contains
//     V - An array of vertices.
//     N - An array of vertex normals.
//     vertex_indices - An array of vertex indices
//                      belonging to the region.
// The output contains
//     signature - Region signature.
void EstimatePrimitive(
	const Eigen::MatrixXd& V,
	const Eigen::MatrixXd& N,
	const std::set<int>& vertex_indices,
	Signature& signature);

#endif