import tensorflow as tf
from .util import *

class Plane:
    @staticmethod
    def get_primitive_name():
        return 'plane'

    @staticmethod
    def compute_parameters(feed_dict, parameters):
        P = feed_dict['P']
        W = feed_dict['W']
        batch_size = tf.shape(P)[0]
        n_points = tf.shape(P)[1]
        n_max_instances = tf.shape(W)[2]
        W_reshaped = tf.reshape(tf.transpose(W, [0, 2, 1]), [batch_size * n_max_instances, n_points]) # BKxN
        P_tiled = tf.reshape(tf.tile(tf.expand_dims(P, axis=1), [1, n_max_instances, 1, 1]), [batch_size * n_max_instances, n_points, 3]) # BKxNx3, important there to match indices in W_reshaped!!!
        n, c = weighted_plane_fitting(P_tiled, W_reshaped) # BKx3
        parameters['plane_n'] = tf.reshape(n, [batch_size, n_max_instances, 3]) # BxKx3
        parameters['plane_c'] = tf.reshape(c, [batch_size, n_max_instances]) # BxK

    @staticmethod
    def compute_residue_loss(parameters, P_gt, matching_indices):
        return Plane.compute_residue_single(
            *adaptor_matching([parameters['plane_n'], parameters['plane_c']], matching_indices), 
            P_gt
        )

    @staticmethod
    def compute_residue_single(n, c, p):
        # n: ...x3, c: ..., p: ...x3
        return tf.square(tf.reduce_sum(p * n, axis=-1) - c)

    def __init__(self, n, c, center=None, x_axis=None, y_axis=None, x_range=[-1, 1],  y_range=[-1, 1]):
        if type(n) is not np.ndarray:
            print('Normal {} needs to be a numpy array!'.format(n))
            raise
        # Plane is defined by {p: n^T p = c}, where the bound is determined by xy_range w.r.t. center
        if center is None:
            center = n * c
        self.n = n / np.linalg.norm(n)
        self.c = c
        self.center = center
        self.x_range = x_range
        self.y_range = y_range

        # parameterize the plane by picking axes
        if x_axis is None or y_axis is None:
            ax_tmp = make_rand_unit_vector()
            self.x_axis = normalized(np.cross(ax_tmp, self.n))
            self.y_axis = normalized(np.cross(self.n, self.x_axis))
        else:
            self.x_axis = x_axis
            self.y_axis = y_axis


