import tensorflow as tf
from .util import *

class Sphere:
    @staticmethod
    def get_primitive_name():
        return 'sphere'

    @staticmethod
    def compute_parameters(feed_dict, parameters):
        P = feed_dict['P']
        W = feed_dict['W']
        batch_size = tf.shape(P)[0]
        n_points = tf.shape(P)[1]
        n_max_primitives = tf.shape(W)[2]
        P = tf.tile(tf.expand_dims(P, axis=1), [1, n_max_primitives, 1, 1]) # BxKxNx3
        W = tf.transpose(W, perm=[0, 2, 1]) # BxKxN
        P = tf.reshape(P, [batch_size * n_max_primitives, n_points, 3]) # BKxNx3
        W = tf.reshape(W, [batch_size * n_max_primitives, n_points]) # BKxN
        fitting_result = weighted_sphere_fitting(P, W)

        parameters['sphere_center'] = tf.reshape(fitting_result['center'], [batch_size, n_max_primitives, 3])
        parameters['sphere_radius_squared'] = tf.reshape(fitting_result['radius_squared'], [batch_size, n_max_primitives])

    @staticmethod
    def compute_residue_loss(parameters, P_gt, matching_indices):
        return Sphere.compute_residue_single(
            *adaptor_matching([parameters['sphere_center'], parameters['sphere_radius_squared']], matching_indices), 
            P_gt
        )

    @staticmethod
    def compute_residue_single(center, radius_squared, p):
        return tf.square(sqrt_safe(tf.reduce_sum(tf.square(p - center), axis=-1)) - sqrt_safe(radius_squared))

    def __init__(self, center, radius):
        self.center = center
        self.radius = radius

