import pickle
import numpy as np
from primitive.plane import Plane
from primitive.sphere import Sphere
from primitive.cone import Cone
from primitive.cylinder import Cylinder

import tensorflow as tf

def log(res, info):
    if res:
        print('%s test success!'%(info))
    else:
        print('%s test fail!'%(info))

def test_compute_parameters():
    data = np.load('../test/test_parameters.npz')
    noises = data['noises']
    plane_points = data['plane_points']
    n = -data['n']
    distance = data['distance']

    sphere_points = data['sphere_points']
    r = data['r']
    c = data['c']

    cylinder_points = data['cylinder_points']
    cylinder_normals = data['cylinder_normals']

    cone_points = data['cone_points']
    cone_normals = data['cone_normals']
    half_angle = data['half_angle']

    plane_error = data['plane_loss']
    sphere_error = data['sphere_loss']
    cylinder_error = data['cylinder_loss']
    cone_error = data['cone_loss']

    member = np.ones((1, 8192, 1)).astype('float32')
    matching_indices = np.zeros((1,1)).astype('int32')
    # Test the implementation
    graph = tf.Graph()
    with graph.as_default():
        P = tf.placeholder(dtype=tf.float32, shape=[1, 8192, 3])
        P_noise = tf.placeholder(dtype=tf.float32, shape=[1, 1, 8192, 3])
        W = tf.placeholder(dtype=tf.float32, shape=[1, 8192, 1])
        N = tf.placeholder(dtype=tf.float32, shape=[1, 8192, 3])
        input_dict = {'P': P, 'W': W, 'normal_per_point': N, }
        parameters = {}
        Plane.compute_parameters(input_dict, parameters)
        Sphere.compute_parameters(input_dict, parameters)
        Cylinder.compute_parameters(input_dict, parameters)
        Cone.compute_parameters(input_dict, parameters)

        PlaneLoss = Plane.compute_residue_loss(parameters, P_noise, matching_indices)
        SphereLoss = Sphere.compute_residue_loss(parameters, P_noise, matching_indices)
        CylinderLoss = Cylinder.compute_residue_loss(parameters, P_noise, matching_indices)
        ConeLoss = Cone.compute_residue_loss(parameters, P_noise, matching_indices)

        sess = tf.Session()
        feed_dict= {
            input_dict['P'] : plane_points,
            input_dict['W'] : member,
            P_noise : (plane_points + noises).reshape((1,1,8192,3))
        }
        plane_loss, plane_n, plane_c = sess.run([PlaneLoss, parameters['plane_n'], parameters['plane_c']], feed_dict)
        if distance * plane_c < 0:
            distance = -distance
            n = -n

        test_plane = True
        if not np.abs(distance - plane_c) < 1e-4:
            test_plane = False
        if not np.linalg.norm(plane_n - n) < 1e-4:
            test_plane = False

        feed_dict= {
            input_dict['P'] : sphere_points,
            input_dict['W'] : member,
            P_noise : (sphere_points + noises).reshape((1,1,8192,3))
        }
        sphere_loss, sphere_center, sphere_radius2 = sess.run([SphereLoss, parameters['sphere_center'], parameters['sphere_radius_squared']], feed_dict)

        test_sphere = True
        if not np.abs(sphere_radius2[0,0] - r * r) < 1e-4:
            test_sphere = False
        if not np.linalg.norm(sphere_center - c) < 1e-4:
            test_sphere = False

        feed_dict= {
            input_dict['P'] : cylinder_points,
            input_dict['W'] : member,
            input_dict['normal_per_point'] : cylinder_normals,
            P_noise : (cylinder_points + noises).reshape((1,1,8192,3))
        }
        cylinder_loss, cylinder_axis, cylinder_center, cylinder_radius2 = sess.run([CylinderLoss, parameters['cylinder_axis'], parameters['cylinder_center'],
            parameters['cylinder_radius_squared']], feed_dict)

        cylinder_axis = cylinder_axis[0,0]
        if np.dot(n, cylinder_axis) < 0:
            n = -n
        
        test_cylinder = True
        if not np.linalg.norm(cylinder_axis - n) < 1e-4:
            test_cylinder = False
        if not np.linalg.norm(np.cross(n, cylinder_center - c)) < 1e-4:
            test_cylinder = False
        if not np.abs(cylinder_radius2 - r * r) < 1e-4:
            test_cylinder = False

        feed_dict= {
            input_dict['P'] : cone_points,
            input_dict['W'] : member,
            input_dict['normal_per_point'] : cone_normals,
            P_noise : (cone_points + noises).reshape((1,1,8192,3))
        }
        cone_loss, cone_apex, cone_axis, cone_half_angle = sess.run([ConeLoss, parameters['cone_apex'], parameters['cone_axis'],
            parameters['cone_half_angle']], feed_dict)

        cone_axis = cone_axis[0,0]
        if np.dot(n, cone_axis) < 0:
            n = -n

        test_cone = True
        if not np.linalg.norm(cone_apex - c) < 1e-4:
            test_cone = False
        if not np.linalg.norm(cone_axis - n) < 1e-4:
            test_cone = False
        if not np.abs(cone_half_angle - half_angle) < 1e-4:
            test_cone = False

        test_plane_err = True
        test_sphere_err = True
        test_cylinder_err = True
        test_cone_err = True

        if not (np.max(np.abs(plane_error - plane_loss)) < 1e-6):
            test_plane_err = False
        if not (np.max(np.abs(sphere_error - sphere_loss)) < 1e-6):
            test_sphere_err = False
        if not (np.max(np.abs(cone_error - cone_loss)) < 1e-6):
            test_cylinder_err = False
        if not (np.max(np.abs(cylinder_error - cylinder_loss)) < 1e-6):
            test_cone_err = False

    log(test_plane, 'Plane parameter')
    log(test_sphere, 'Sphere parameter')
    log(test_cylinder, 'Cylinder parameter')
    log(test_cone, 'Cone parameter')

    log(test_plane_err, 'Plane distance')
    log(test_sphere_err, 'Sphere distance')
    log(test_cylinder_err, 'Cylinder distance')
    log(test_cone_err, 'Cone distance')

if __name__ == "__main__":
    test_compute_parameters()